#include "user_main.h"

#define MY_SSID					"SSID"
#define MY_PASSWORD				"Password"

#define UDP_PORT				8888

#define MAX_NET_CONN			20

os_event_t procTaskQueue[TASK_COUNT];
uint8_t dbgChan;
uint16_t loopCount;
uint16_t state;
uint8_t connected;
int16_t handle = -1;
int16_t stat;

NetConn_t *netConn[MAX_NET_CONN];

void ZB_FUNC_ATTR_NORMAL
user_init(void)
{
	struct station_config config;

	state = 0;
	loopCount = 0;

	// initialize the debug output channel
	dbgInit(DEBUG_UART, 115200);
	dbgCRLF();

	// set to station mode
	if (wifi_get_opmode() != 1)
		wifi_set_opmode(1);
	strcpy(config.ssid, MY_SSID);
	strcpy(config.password, MY_PASSWORD);
	config.bssid_set = 0;
	wifi_station_set_config(&config);
    wifi_station_dhcpc_start();
	wifi_station_set_auto_connect(1);

	system_os_task(mainTask, 0, procTaskQueue, TASK_COUNT);
	system_os_post(0, 0, 0);
}

void ZB_FUNC_ATTR_NORMAL
mainTask(os_event_t *events)
{
	if (++loopCount >= 1000)
	{
		loopCount = 0;
		if (state < 10)
		{
			uint8_t wifiStat = wifi_station_get_connect_status();
			dbgStr("state = ");
			dbgWordDec(state);
			dbgStr(", WiFi status is ");
			dbgWordDec(wifiStat);
			dbgCRLF();
			if (!connected & (wifiStat == 5))
			{
				uint32_t addr = wifiGetIP();
				dspIP("connected, IP = ", addr);
				dbgCRLF();
				connected = 1;
				if (addr)
				{
					handle = netOpen();
					if (handle >= 0)
					{
						stat = netSetCallback(handle, (NetCallback_t)netCallback);
						stat = netListen(handle, UDP_PORT, PROTO_UDP);
						dbgStr("Net.Listen() returns ");
						dbgWordDec(stat);
						dbgCRLF();
					}
				}
			}
		}
		if (++state > 1000)
			state = 1000;
	}
    os_delay_us(1000);
	system_os_post(0, 0, 0);
}

void ZB_FUNC_ATTR_NORMAL
netCallback(uint8_t cbtype, int16_t handle, uint8_t *data, uint16_t cnt)
{
	dbgStr("chan ");
	dbgWordDec(handle);
	dbgChar(' ');

	switch(cbtype)
	{
	case NET_CONNECT:
		dbgStr("cb: connect\r\n");
		break;

	case NET_DISCONNECT:
		dbgStr("cb: disconnect\r\n");
		break;

	case NET_RX_DATA:
		dbgStr("cb: receive ");
		dbgWordDec(cnt);
		dbgCRLF();
		dspData(data, cnt);

		// send a reply to the client
		netSendStr(handle, "Goodbye, world!");
		break;

	case NET_TX_DATA:
		dbgStr("cb: send\r\n");
		netClose(handle);
		break;

	default:
		dbgStr("cb: unknown (");
		dbgWordDec(cbtype);
		dbgStr(")\r\n");
		break;
	}
}

int16_t ZB_FUNC_ATTR_NORMAL
netOpen(void)
{
	int idx;
	for (idx = 0; idx < MAX_NET_CONN; idx++)
	{
		if (netConn[idx] == NULL)
		{
			// attempt to allocate a control structure
			NetConn_t *ncp = memZalloc(sizeof(*ncp));
			if (ncp == NULL)
				return(ERR_MEM);

			// store the pointer to the control structure and initialize some elements
			netConn[idx] = ncp;
			ncp->chan = idx;
			ncp->timeout = DEFAULT_TIMEOUT;
			return(idx);
		}
	}
	return(ERR_ERR);	
}

int16_t ZB_FUNC_ATTR_NORMAL
netClose(int16_t chan)
{
	int16_t stat = ERR_OK;
	NetConn_t *ncp;

	if ((chan < 0) || (chan >= MAX_NET_CONN))
		return(ERR_ARG);
	if ((ncp = netConn[chan]) == NULL)
		return(ERR_CLSD);
	if (ncp->conn != NULL)
	{
		NetCallback_t cb = NULL;
		switch (ncp->conn->type)
		{
#ifdef THIS_IS_NOT_NEEDED
		case ESPCONN_TCP:
			if (ncp->flags & CONNECTED)
			{
				stat = espconn_disconnect(ncp->conn);
				cb = ncp->callback;
			}
			else
				stat = ERR_OK;
			break;
#endif

		case ESPCONN_UDP:
			if (ncp->flags & CONNECTED)
				cb = ncp->callback;
			stat = ERR_OK;
			break;

		default:
			// this shouldn't happen
			break;
		}

		// deallocate connection information
		if ((ncp->flags & CONN_ALLOCATED) && !(ncp->flags & CLIENT_MODE))
		{
			stat = espconn_delete(ncp->conn);
// N.B: this should not happen for the handle associated with the remote host
dbgStr("espconn_delete() returns ");
dbgInt(stat);
dbgCRLF();
		}

		if (ncp->flags & CONN_ALLOCATED)
		{
// N.B: this should not happen for the handle associated with the remote host
dbgStr("deallocating pcb and connection\r\n");
			// deallocate the TCP/UDP protocol block
			if (ncp->conn->proto.tcp != NULL)
				memFree(ncp->conn->proto.tcp);

			// deallocate the connection
			memFree(ncp->conn);
		}

		// kill the conection
dbgStr("killing the connection handle\r\n");
		ncp->conn = NULL;

		if (cb)
			(*cb)(NET_DISCONNECT, chan, NULL, 0);
	}

//N.B.: if this line is commented out the program works (but has a resource leak).
//      With it uncommented, the app fails to respond after the first datagram is received.
	// free the control block
//	memFree(ncp);

	netConn[chan] = NULL;
	return(stat);
}

int16_t ZB_FUNC_ATTR_NORMAL
netSetCallback(int16_t chan, NetCallback_t cb)
{
	NetConn_t *ncp;

	if ((chan < 0) || (chan >= MAX_NET_CONN))
		return(ERR_ARG);
	if ((ncp = netConn[chan]) == NULL)
		return(ERR_CLSD);
	ncp->callback = cb;
	return(ERR_OK);
}

int16_t ZB_FUNC_ATTR_NORMAL
netListen(int16_t chan, uint32_t port, uint8_t proto)
{
	int16_t stat;
	IPAddress_t ip;
	NetConn_t *ncp;
	struct espconn *conn;
	esp_tcp *pcb;
	enum espconn_type type;

	// check for a valid channel
	if ((chan < 0) || (chan >= MAX_NET_CONN) || (proto > PROTO_MAX))
		return(ERR_ARG);

	// check that the channel is open but not yet connected
	if ((ncp = netConn[chan]) == NULL)
		return(ERR_CLSD);
	if (ncp->conn && (ncp->conn->type || ncp->conn->proto.tcp))
		return(ERR_ISCONN);

	// allocate the protocol control block
	type = ((proto & 0x01) ? ESPCONN_UDP : ESPCONN_TCP);
	if ((pcb = memZalloc((type == ESPCONN_TCP) ? sizeof(esp_tcp) : sizeof(esp_udp))) == NULL)
		return(ERR_MEM);
	if ((conn = memZalloc(sizeof(*conn))) == NULL)
	{
		memFree(pcb);
		return(ERR_MEM);
	}

	// prepare for a connection
	ncp->flags = CONN_ALLOCATED;
	ncp->conn = conn;
	conn->proto.tcp = pcb;
	conn->type = type;
	conn->state = ESPCONN_NONE;
	conn->reverse = ncp;

	// prepare the addresses and ports
	ip.w = wifiGetIP();
	memcpy(pcb->local_ip, &ip.b, 4);
	pcb->local_port = port;	
	ip.w = 0;
	memcpy(pcb->remote_ip, &ip.b, 4);
	pcb->remote_port = 0;

	switch (type)
	{
#ifdef THIS_IS_NOT_NEEDED
	case ESPCONN_TCP:
		// prepare a TCP server connection
		stat = espconn_regist_connectcb(ncp->conn, tcpConnectCB);
		stat = espconn_accept(ncp->conn);
		stat = espconn_regist_time(ncp->conn, DEF_SERVER_TIMEOUT, 0);
		break;
#endif

	case ESPCONN_UDP:
		// prepare a UDP server connection
		stat = espconn_regist_recvcb(ncp->conn, udpReceiveCB);
		stat = espconn_regist_sentcb(ncp->conn, netSendDoneCB);
		stat = espconn_create(ncp->conn);
		break;

	default:
		// this should never happen
		stat = ERR_ERR;
		break;
	}
	return(stat);
}

void ZB_FUNC_ATTR_NORMAL
udpReceiveCB(void *arg, char *data, uint16_t cnt)
{
	int hand;
	NetConn_t *ncpOld;
	NetConn_t *ncpNew;
	struct espconn *ecp = (struct espconn *)arg;

	if ((ecp != NULL) && ((ncpOld = ecp->reverse) != NULL) &&
			((hand = netOpen()) >= 0) && ((ncpNew = netConn[hand]) != NULL)) 
	{
		// populate the new connection handle
		ecp->reverse = ncpNew;
		ncpNew->conn = ecp;
		ncpNew->callback = ncpOld->callback;
		ncpNew->flags = CONNECTED;

		if (ncpNew->callback != NULL)
			(*ncpNew->callback)(NET_RX_DATA, hand, (uint8_t *)data, cnt);
	}

}

void ZB_FUNC_ATTR_NORMAL
netSendDoneCB(void *arg)
{
	NetConn_t *ncp;
	struct espconn *ecp = (struct espconn *)arg;

	if ((ecp != NULL) && ((ncp = ecp->reverse) != NULL))
	{
		if (ncp->callback != NULL)
			(*ncp->callback)(NET_TX_DATA, ncp->chan, NULL, 0);
	}
}

int16_t ZB_FUNC_ATTR_NORMAL
netSend(NetConn_t *ncp, const uint8_t *data, uint16_t cnt)
{
	int stat = ERR_OK;

	if ((ncp == NULL) || (ncp->conn == NULL) || (data == NULL))
		return(ERR_ARG);
	if (cnt == 0)
		return(ERR_OK);

	// the channel is connected, send the data
	ncp->flags |= AWAITING_TRANSMISSION;
	return(espconn_sent(ncp->conn, (uint8_t *)data, cnt));
}

int16_t ZB_FUNC_ATTR_NORMAL
netSendStr(int16_t chan, const uint8_t *str)
{
	if (str != NULL)
		return(netSendData(chan, str, strlen((char *)str)));
	return(ERR_OK);
}

int16_t ZB_FUNC_ATTR_NORMAL
netSendData(int16_t chan, const uint8_t *data, uint16_t cnt)
{
	NetConn_t *ncp;

	if ((chan < 0) || (chan >= MAX_NET_CONN) || (data == NULL))
		return(ERR_ARG);
	if ((ncp = netConn[chan]) == NULL)
		return(ERR_CLSD);
	if ((ncp->conn->type == ESPCONN_TCP) && !(ncp->flags & CONNECTED))
		return(ERR_CONN);
	return(netSend(ncp, data, cnt));
}

uint32_t ZB_FUNC_ATTR_NORMAL
wifiGetIP(void)
{
	struct ip_info ipInfo;
	wifi_get_ip_info(0, &ipInfo);
	return(ipInfo.ip.addr);
}

void ZB_FUNC_ATTR_NORMAL
dbgInit(uint8_t chan, uint32_t baud)
{
	// set the debug channel
	if (chan > 1)
		chan = 1;
	dbgChan = chan;

	if (baud != 0)
	{
		uint32_t flags;
		uint32_t mask;
		if (dbgChan == 1)
			PIN_FUNC_SELECT(PERIPHS_IO_MUX_GPIO2_U, FUNC_U1TXD_BK);
		else
		{
			PIN_PULLUP_DIS(PERIPHS_IO_MUX_U0TXD_U);
			PIN_FUNC_SELECT(PERIPHS_IO_MUX_U0TXD_U, FUNC_U0TXD);
		}
		uart_div_modify(dbgChan, UART_CLK_FREQ / baud);

		mask = (UART_STOP_BIT_NUM << UART_STOP_BIT_NUM_S) |
				(UART_BIT_NUM << UART_BIT_NUM_S) | UART_PARITY_EN | UART_PARITY;
		flags = READ_PERI_REG(UART_CONF0(dbgChan)) & ~mask;

		// default to N81 operation
		flags |= (NONE_BITS & (UART_PARITY_EN | UART_PARITY));
		flags |= (EIGHT_BITS & UART_BIT_NUM) << UART_BIT_NUM_S;
		flags |= (ONE_STOP_BIT & UART_STOP_BIT_NUM) << UART_STOP_BIT_NUM_S;
		WRITE_PERI_REG(UART_CONF0(dbgChan), flags);

		// clear rx and tx fifo, not ready
		SET_PERI_REG_MASK(UART_CONF0(dbgChan), UART_RXFIFO_RST | UART_TXFIFO_RST);
		CLEAR_PERI_REG_MASK(UART_CONF0(dbgChan), UART_RXFIFO_RST | UART_TXFIFO_RST);

		//clear all pending interrupts and disable
		WRITE_PERI_REG(UART_INT_CLR(dbgChan), 0xffff);
		CLEAR_PERI_REG_MASK(UART_INT_ENA(dbgChan), 0xff);
	}
}

void ZB_FUNC_ATTR_NORMAL
dbgChar(uint8_t c)
{
	while (1)
	{
		uint32_t fifoCnt = READ_PERI_REG(UART_STATUS(dbgChan)) & (UART_TXFIFO_CNT << UART_TXFIFO_CNT_S);
		if (((fifoCnt >> UART_TXFIFO_CNT_S) & UART_TXFIFO_CNT) < 126)
			break;
	}
	WRITE_PERI_REG(UART_FIFO(dbgChan), c);
}

void ZB_FUNC_ATTR_NORMAL
dbgCRLF(void)
{
	dbgChar(0x0d);
	dbgChar(0x0a);
}

void ZB_FUNC_ATTR_NORMAL
dbgByteDec(uint8_t val)
{
	dbgWordDec(val);
}

void ZB_FUNC_ATTR_NORMAL
dbgWordDec(uint16_t val)
{
	if (val >= 10)
		dbgWordDec(val / 10);
	dbgChar((uint8_t)(val % 10) + '0');
}

void ZB_FUNC_ATTR_NORMAL
dbgInt(int16_t val)
{
	if (val < 0)
	{
		dbgChar('-');
		val = -val;
	}
	dbgWordDec(val);
}

void ZB_FUNC_ATTR_NORMAL
dbgStr(const char *str)
{
	if (str != NULL)
	{
		char c;
		while ((c = *str++) != '\0')
			dbgChar(c);
	}
}

void ZB_FUNC_ATTR_NORMAL
dbgStrN(const char *str, uint16_t cnt)
{
	if (str != NULL)
	{
		while (cnt--)
			dbgChar(*str++);
	}
}

void * ZB_FUNC_ATTR_NORMAL
memAlloc(size_t size)
{
	return((size == 0) ? NULL : os_malloc(ADJUST_ALLOC_BLOCK(size)));
}

void * ZB_FUNC_ATTR_NORMAL
memZalloc(size_t size)
{
	return((size == 0) ? NULL : os_zalloc(ADJUST_ALLOC_BLOCK(size)));
}

void ZB_FUNC_ATTR_NORMAL
memFree(void *p)
{
	if (p != NULL)
	   	os_free(p);
}

void ZB_FUNC_ATTR_NORMAL
dspData(uint8_t *data, uint16_t cnt)
{
	dbgStr("------------------\r\n");
	while (cnt--)
	{
		uint8_t c;
		c = *data++;
		if (c == '\n')
			dbgCRLF();
		else
			dbgChar(c);
	}
	dbgStr("------------------\r\n");
}

void ZB_FUNC_ATTR_NORMAL
dspIP(const char *caption, uint32_t ip)
{
	int16_t i;
	IPAddress_t addr;

	dbgStr(caption);
	addr.w = ip;
	for (i = 0; i < 4; i++)
	{
		dbgByteDec(addr.b[i]);
		if (i < 3)
			dbgChar('.');
	}
}

void ZB_FUNC_ATTR_NORMAL
user_rf_pre_init(void)
{
}
